'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 0.1.10

version: 0.1.10

full_version: 0.1.10

git_revision: 499a50af6474eaa932c572ff9aa0633865ba11c5

release: True

'''
short_version: str = '0.1.10'
version: str = '0.1.10'
full_version: str = '0.1.10'
git_revision: str = '499a50af6474eaa932c572ff9aa0633865ba11c5'
release: bool = True
if not release:
    version = full_version



import numpy as np
import pandas as pd

from scipy.optimize import minimize

from functools import partial
import warnings

import datetime

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.linear_model import LinearRegression

from typing import Union
from abc import abstractmethod, ABC

import numbers

class NaivePredictor(BaseEstimator, RegressorMixin):
    """
    Naive predictor class to output a column of features as the predictions themselves.
    It expects only one column in the input matrix for which predictions are to be made.
    This class would typically be used in the final stage of a `scikit-learn` pipeline,
    following either `FeatureAverager` or `ZnScoreAverager` stages.
    """

    def fit(self, X, y=None):
        return self

    def predict(self, X):
        return np.array(X)


class BaseWeightedRegressor(BaseEstimator, RegressorMixin, ABC):
    def fit(self, X: pd.DataFrame, y: Union[pd.DataFrame, pd.Series]):
        """
        Fit method to fit the underlying model with weighted samples, as passed
        into the constructor.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.

        :return <BaseWeightedRegressor>
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame) and y.shape[1] != 1:
            raise ValueError(
                "The target dataframe must have only one column. If used as part of "
                "an sklearn pipeline, ensure that previous steps return a pandas "
                "series or dataframe."
            )

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't "
                "match."
            )

        # Fit
        self.sample_weights = self._calculate_sample_weights(y)
        self.model.fit(X, y, sample_weight=self.sample_weights)
        if hasattr(self.model, "coef_"):
            self.coef_ = self.model.coef_
        if hasattr(self.model, "intercept_"):
            self.intercept_ = self.model.intercept_

        return self

    def predict(self, X: pd.DataFrame):
        """
        Predict method to make model predictions on the input feature matrix X based on
        the previously fit model.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <np.ndarray>: Numpy array of predictions.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the SignWeightedLinearRegression must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")

        # Predict
        return self.model.predict(X)

    @abstractmethod
    def _calculate_sample_weights(self, y: Union[pd.DataFrame, pd.Series]):
        """
        Abstract method for calculating sample weights based on the target variable `y`.
        """
        pass


class SignWeightedRegressor(BaseWeightedRegressor):

    def __init__(self, model: BaseEstimator):
        """
        Custom class to create a weighted regressor, with the sample weights
        chosen by inverse frequency of the label's sign in the training set.

        :param <BaseEstimator> model: The underlying model to be trained with weighted samples.

        NOTE: By weighting the contribution of different training samples based on the
        sign of the label, the model is encouraged to learn equally from both positive and negative return samples,
        irrespective of class imbalance. If there are more positive targets than negative
        targets in the training set, then the negative target samples are given a higher
        weight in the model training process. The opposite is true if there are more
        negative targets than positive targets.
        """
        self.model = model

    def _calculate_sample_weights(self, targets: Union[pd.DataFrame, pd.Series]):
        """
        Private helper method to calculate the sample weights, chosen by inverse frequency
        of the label's sign in the training set.

        :param <Union[pd.DataFrame, pd.Series]> targets: Pandas series or dataframe of targets.

        :return <tuple[np.ndarray, float, float]>: Tuple of sample weights, positive weight and negative weight.
        """
        pos_sum = np.sum(targets >= 0)
        neg_sum = np.sum(targets < 0)

        pos_weight = len(targets) / (2 * pos_sum) if pos_sum > 0 else 0
        neg_weight = len(targets) / (2 * neg_sum) if neg_sum > 0 else 0

        sample_weights = np.where(targets >= 0, pos_weight, neg_weight)
        return sample_weights


class TimeWeightedRegressor(BaseWeightedRegressor):
    def __init__(self, half_life: Union[float, int], model: BaseEstimator):
        """
        Custom class to create a weighted regressor, where the training sample
        weights exponentially decay by sample recency, given a prescribed half_life.

        :param <Union[float, int]> half_life: The number of time periods in units of the native data frequency for the weight attributed to the most recent sample (one) to decay by half.
        :param <BaseEstimator> model: The underlying model to be trained with weighted samples.

        NOTE: By weighting the contribution of different training samples based on the
        observation date, the model is encouraged to prioritise newer information.
        The half-life denotes the number of time periods in units of the native data
        frequency for the weight attributed to the most recent sample (one) to decay by half.
        """
        if not isinstance(half_life, (float, int)):
            raise TypeError("half_life must be a float or an integer.")
        if half_life <= 1:
            raise ValueError("The half-life must be greater than 1.")
        self.half_life = half_life
        self.model = model

    def _calculate_sample_weights(self, targets: Union[pd.DataFrame, pd.Series]):
        """
        Private helper method to calculate the sample weights, chosen by an exponentially
        decaying function of the sample recency.

        :param <Union[pd.DataFrame, pd.Series]> targets: Pandas series or dataframe of targets.

        :return <np.ndarray>: Numpy array of sample weights.
        """

        dates = sorted(targets.index.get_level_values(1).unique(), reverse=True)
        num_dates = len(dates)
        weights = np.power(2, -np.arange(num_dates) / self.half_life)

        weight_map = dict(zip(dates, weights))
        self.sample_weights = targets.index.get_level_values(1).map(weight_map).to_numpy()

        return self.sample_weights


class SignWeightedLinearRegression(SignWeightedRegressor):
    def __init__(
        self,
        fit_intercept: bool = True,
        copy_X: bool = True,
        n_jobs: int = None,
        positive: bool = False,
    ):
        """
        Custom class to create a WLS linear regression model, with the sample weights
        chosen by inverse frequency of the label's sign in the training set.

        :param <bool> fit_intercept: Whether to calculate the intercept for this model. If set to False, no intercept will be used in calculations (i.e. data is expected to be centered).
        :param <bool> copy_X: If True, X will be copied; else, it may be overwritten.
        :param <int> n_jobs: The number of jobs to use for the computation.
        :param <bool> positive: When set to True, forces the coefficients to be positive. This option is only supported for dense arrays.

        NOTE: By weighting the contribution of different training samples based on the
        sign of the label, the model is encouraged to learn equally from both positive and negative return samples,
        irrespective of class imbalance. If there are more positive targets than negative
        targets in the training set, then the negative target samples are given a higher
        weight in the model training process. The opposite is true if there are more
        negative targets than positive targets.
        """

        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(copy_X, bool):
            raise TypeError("copy_X must be a boolean.")
        if not isinstance(n_jobs, int) and n_jobs != None:
            raise TypeError("n_jobs must be an integer or None.")
        if n_jobs is not None:
            if n_jobs < -1 or n_jobs == 0:
                raise ValueError("n_jobs must be a positive integer or -1.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")

        self.fit_intercept = fit_intercept
        self.copy_X = copy_X
        self.n_jobs = n_jobs
        self.positive = positive
        model = LinearRegression(
            fit_intercept=self.fit_intercept,
            copy_X=self.copy_X,
            n_jobs=self.n_jobs,
            positive=self.positive,
        )
        self.coef_ = None 
        self.intercept_ = None
        super().__init__(model)

    def set_params(self, **params):
        super().set_params(**params)
        if 'fit_intercept' in params or 'positive' in params:
            # Re-initialize the LinearRegression instance with updated parameters
            self.model = LinearRegression(
                fit_intercept=self.fit_intercept,
                positive=self.positive
            )

        return self

class TimeWeightedLinearRegression(TimeWeightedRegressor):
    def __init__(
        self,
        fit_intercept: bool = True,
        copy_X: bool = True,
        n_jobs: int = None,
        positive: bool = False,
        half_life: Union[float, int] = 21 * 12,
    ):
        """
        Custom class to create a WLS linear regression model, where the training sample
        weights exponentially decay by sample recency, given a prescribed half_life.

        :param <bool> fit_intercept: Whether to calculate the intercept for this model. If set to False, no intercept will be used in calculations (i.e. data is expected to be centered).
        :param <bool> copy_X: If True, X will be copied; else, it may be overwritten.
        :param <int> n_jobs: The number of jobs to use for the computation.
        :param <bool> positive: When set to True, forces the coefficients to be positive. This option is only supported for dense arrays.
        :param <Union[float, int]> half_life: The number of time periods in units of the native data frequency for the weight attributed to the most recent sample (one) to decay by half.
        """
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(copy_X, bool):
            raise TypeError("copy_X must be a boolean.")
        if not isinstance(n_jobs, int) and n_jobs != None:
            raise TypeError("n_jobs must be an integer or None.")
        if n_jobs is not None:
            if n_jobs < -1 or n_jobs == 0:
                raise ValueError("n_jobs must be a positive integer or -1.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")

        self.fit_intercept = fit_intercept
        self.copy_X = copy_X
        self.n_jobs = n_jobs
        self.positive = positive
        model = LinearRegression(
            fit_intercept=self.fit_intercept,
            copy_X=self.copy_X,
            n_jobs=self.n_jobs,
            positive=self.positive,
        )
        self.coef_ = None
        self.intercept_ = None
        super().__init__(half_life=half_life, model=model)

    def set_params(self, **params):
        super().set_params(**params)
        if 'fit_intercept' in params or 'positive' in params:
            # Re-initialize the LinearRegression instance with updated parameters
            self.model = LinearRegression(
                fit_intercept=self.fit_intercept,
                positive=self.positive
            )

        return self

class SignWeightedLADRegressor(SignWeightedRegressor):
    def __init__(
        self,
        fit_intercept: bool = True,
        positive: bool = False,
    ):
        """
        Custom class to create a weighted LAD linear regression model, with the sample weights
        chosen by inverse frequency of the label's sign in the training set.

        :param <bool> fit_intercept: Whether to calculate the intercept for this model. If set to False, no intercept will be used in calculation.
        :param <bool> positive: When set to True, forces the coefficients to be positive.

        NOTE: By weighting the contribution of different training samples based on the
        sign of the label, the model is encouraged to learn equally from both positive and negative return samples,
        irrespective of class imbalance. If there are more positive targets than negative
        targets in the training set, then the negative target samples are given a higher
        weight in the model training process. The opposite is true if there are more
        negative targets than positive targets.
        """

        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")

        self.fit_intercept = fit_intercept
        self.positive = positive
        model = LADRegressor(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
        )
        self.coef_ = None
        self.intercept_ = None
        super().__init__(model)

    def set_params(self, **params):
        super().set_params(**params)
        if 'fit_intercept' in params or 'positive' in params:
            # Re-initialize the LinearRegression instance with updated parameters
            self.model = LADRegressor(
                fit_intercept=self.fit_intercept,
                positive=self.positive
            )

        return self


class TimeWeightedLADRegressor(TimeWeightedRegressor):
    def __init__(
        self,
        fit_intercept: bool = True,
        positive: bool = False,
        half_life: Union[float, int] = 21 * 12,
    ):
        """
        Custom class to create a weighted LAD linear regression model, where the training sample
        weights exponentially decay by sample recency, given a prescribed half_life.

        :param <bool> fit_intercept: Whether to calculate the intercept for this model. If set to False, no intercept will be used in calculations.
        :param <bool> positive: When set to True, forces the coefficients to be positive. This option is only supported for dense arrays.
        :param <Union[float, int]> half_life: The number of time periods in units of the native data frequency for the weight attributed to the most recent sample (one) to decay by half.
        """
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")

        self.fit_intercept = fit_intercept
        self.positive = positive
        model = LADRegressor(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
        )
        self.coef_ = None
        self.intercept_ = None
        super().__init__(half_life=half_life, model=model)

    def set_params(self, **params):
        super().set_params(**params)
        if 'fit_intercept' in params or 'positive' in params:
            # Re-initialize the LinearRegression instance with updated parameters
            self.model = LADRegressor(
                fit_intercept=self.fit_intercept,
                positive=self.positive
            )

        return self

class LADRegressor(BaseEstimator, RegressorMixin):
    def __init__(self, fit_intercept=True, positive=False, tol=None, ):
        """
        Custom class to create a linear regression model with model fit determined
        by minimising L1 (absolute) loss.

        :param <bool> fit_intercept: Whether to calculate the intercept for this model.
            If set to False, no intercept will be used in calculations.
        :param <bool> positive: If True, the optimisation is restricted so that model
            weights (other than the intercept) are greater or equal to zero.
            This is a way of incorporating prior knowledge about the relationship between
            model features and corresponding targets into the model fit.
        :param <float> tol: The tolerance for termination. Default is None.

        NOTE: Standard OLS linear regression models are fit by minimising the residual
            sum of squares. Our implemented (L)east (A)bsolute (D)eviation regression
            model fits a linear model by minimising the residual absolute deviations.
            Consequently, the model fit is more robust to outliers than an OLS fit.
        """
        # Checks
        if not isinstance(fit_intercept, bool):
            raise TypeError("'fit_intercept' must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("'positive' must be a boolean.")
        if tol is not None and not isinstance(tol, (int, float)):
            raise TypeError("'tol' must be a float or int.")
        if tol is not None and tol <= 0:
            raise ValueError("'tol' must be a positive number.")

        # Initialise
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.tol = tol
        self.coef_ = None
        self.intercept_ = None

    def fit(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        sample_weight: np.ndarray = None,
    ):
        """
        Fit method to fit a LAD linear regression model.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        :param <np.ndarray> sample_weight: Numpy array of sample weights to create a
            weighted LAD regression model. Default is None for equal sample weights.

        :return <LADRegressor>
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the LADRegressor must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector for the LADRegressor must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The target dataframe must have only one column. If used as part of "
                    "an sklearn pipeline, ensure that previous steps return a pandas "
                    "series or dataframe."
                )

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't "
                "match."
            )

        if sample_weight is not None:
            if not isinstance(sample_weight, np.ndarray):
                try:
                    sample_weight = sample_weight.to_numpy().flatten()
                except Exception as e:
                    try:
                        sample_weight = np.array(sample_weight).flatten()
                    except Exception as e:
                        raise TypeError(
                            "The sample weights must be contained within a numpy array."
                        )
            if sample_weight.ndim != 1:
                raise ValueError("The sample weights must be a 1D numpy array.")
            for w in sample_weight:
                if not isinstance(w, numbers.Number):
                    raise TypeError(
                        "All elements of the sample weights must be numeric."
                    )
            if len(sample_weight) != X.shape[0]:
                raise ValueError(
                    "The number of sample weights must match the number of samples in the input feature matrix."
                )

        # Fit
        X = X.copy()
        y = y.copy()

        if self.fit_intercept:
            X.insert(0, "intercept", 1)

        n_cols = X.shape[1]

        if self.positive:
            if self.fit_intercept:
                bounds = [(None, None)] + [(0, None) for _ in range(n_cols - 1)]
            else:
                bounds = [(0, None) for _ in range(n_cols)]
        else:
            bounds = [(None, None) for _ in range(n_cols)]

        # optimisation
        init_weights = np.zeros(n_cols)
        optim_results = minimize(
            fun=partial(
                self._l1_loss,
                X=X,
                y=y,
                sample_weight=sample_weight,
            ),
            x0=init_weights,
            method="SLSQP", # TODO: make this an option in the constructor.
            bounds=bounds,
            tol=self.tol,
        )

        if not optim_results.success:
            self.intercept_ = None
            self.coef_ = None
            warnings.warn(
                "LADRegressor optimisation failed. Setting intercept and coefficients to None.",
                RuntimeWarning,
            )
            return self

        if self.fit_intercept:
            # Then store the intercept and feature weights
            self.intercept_ = optim_results.x[0]
            self.coef_ = optim_results.x[1:]
        else:
            self.intercept_ = None
            self.coef_ = optim_results.x

        return self

    def predict(self, X: pd.DataFrame):
        """
        Predict method to make model predictions on the input feature matrix X based on
        the previously fit model.

        :param <pd.DataFrame> X: Pandas dataframe of input features.

        :return <np.ndarray>: Numpy array of predictions.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the LADRegressor must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")

        # Predict
        if self.intercept_ is None and self.coef_ is None:
            warnings.warn(
                "LADRegressor model fit failed. Predicting a zero signal.",
                RuntimeWarning,
            )
            return pd.Series(0, index=X.index)

        X = X.copy()
        if self.fit_intercept:
            return (X.dot(self.coef_) + self.intercept_).values
        else:
            return X.dot(self.coef_).values

    def _l1_loss(
        self,
        weights: np.ndarray,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        sample_weight: np.ndarray = None,
    ):
        """
        Private helper method to determine the mean training L1 loss induced by 'weights'.

        :param <np.ndarray> weights: LADRegressor model coefficients to be optimised.
        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        :param <np.ndarray> sample_weight: Numpy array of sample weights to create a
            weighted LAD regression model. Default is None for equal sample weights.

        :return <float>: Training loss induced by 'weights'.
        """
        # Checks
        if not isinstance(weights, np.ndarray):
            raise TypeError("The weights must be contained within a numpy array.")

        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the LADRegressor must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector for the LADRegressor must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The target dataframe must have only one column. If used as part of "
                    "an sklearn pipeline, ensure that previous steps return a pandas "
                    "series or dataframe."
                )

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't "
                "match."
            )

        if sample_weight is None:
            sample_weight = np.ones(X.shape[0])

        if not isinstance(sample_weight, np.ndarray):
            raise TypeError(
                "The sample weights must be contained within a numpy array."
            )
        if sample_weight.ndim != 1:
            raise ValueError("The sample weights must be a 1D numpy array.")
        if len(sample_weight) != X.shape[0]:
            raise ValueError(
                "The number of sample weights must match the number of samples in the input feature matrix."
            )

        if isinstance(y, pd.DataFrame):
            raw_residuals = y.iloc[:, 0] - X.dot(weights)
        else:  # y is a series
            raw_residuals = y - X.dot(weights)
        abs_residuals = np.abs(raw_residuals)
        weighted_abs_residuals = abs_residuals * sample_weight

        return np.mean(weighted_abs_residuals)


if __name__ == "__main__":
    from macrosynergy.management import make_qdf
    import macrosynergy.management as msm
    import matplotlib.pyplot as plt

    from sklearn.linear_model import LinearRegression

    np.random.seed(1)

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2013-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2010-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2010-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2010-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2010-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd.drop(columns=["XR"])
    y = dfd["XR"]

    # OLS Linear Regression
    lm = LinearRegression()
    lm.fit(X, y)
    lm_preds = lm.predict(X)
    # Sign-Weighted Linear Regression
    swlm = SignWeightedLinearRegression()
    swlm.fit(X, y)
    swlm_preds = swlm.predict(X)
    # Time-Weighted Linear Regression
    twlm = TimeWeightedLinearRegression(half_life=12 * 21)  # One year half life
    twlm.fit(X, y)
    twlm_preds = twlm.predict(X)

    # Plot results
    plt.title(
        "OLS vs Sign-Weighted vs Time-Weighted Linear Regression on mock random data"
    )
    plt.plot(lm_preds, label="OLS", color="tab:blue", alpha=0.5)
    plt.plot(swlm_preds, label="Sign-Weighted", color="tab:orange", alpha=0.5)
    plt.plot(twlm_preds, label="Time-Weighted", color="tab:green", alpha=0.5)

    plt.legend()
    plt.show()

    # Plot histogram of predictions
    fig, ax = plt.subplots(ncols=3)
    ax[0].hist(lm_preds, bins=200, color="tab:blue")
    ax[0].set_title("OLS")
    ax[1].hist(swlm_preds, bins=200, color="tab:orange")
    ax[1].set_title("Sign-Weighted")
    ax[2].hist(twlm_preds, bins=200, color="tab:green")
    ax[2].set_title("Time-Weighted")

    plt.suptitle("Histograms of predictions")
    plt.tight_layout()
    plt.show()



"""
A subclass inheriting from `macrosynergy.visuals.plotter.Plotter`,
designed to plot time series data on a line plot.
"""

import pandas as pd
from typing import List, Dict, Tuple, Optional
import matplotlib.pyplot as plt

from macrosynergy.visuals.plotter import Plotter
from macrosynergy.management.types import Numeric, NoneType

from macrosynergy.management.simulate import make_test_df


class LinePlot(Plotter):
    """
    Class for plotting time series data on a line plot.
    Inherits from `macrosynergy.visuals.plotter.Plotter`.

    Parameters
    :param <pd.DataFrame> df: A DataFrame with the following columns:
        'cid', 'xcat', 'real_date', and at least one metric from -
        'value', 'grading', 'eop_lag', or 'mop_lag'.
    :param <List[str]> cids: A list of cids to select from the DataFrame.
        If None, all cids are selected.
    :param <List[str]> xcats: A list of xcats to select from the DataFrame.
        If None, all xcats are selected.
    :param <List[str]> metrics: A list of metrics to select from the DataFrame.
        If None, all metrics are selected.
    :param <str> start_date: ISO-8601 formatted date. Select data from
        this date onwards. If None, all dates are selected.
    :param <str> end_date: ISO-8601 formatted date. Select data up to
        and including this date. If None, all dates are selected.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
        intersect: bool = False,
        tickers: Optional[List[str]] = None,
        blacklist: Optional[Dict[str, List[str]]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            df=df,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            intersect=intersect,
            tickers=tickers,
            blacklist=blacklist,
            start=start,
            end=end,
            *args,
            **kwargs,
        )

    def plot(
        self,
        # df/plot args
        metric: str = "value",
        compare_series: Optional[str] = None,
        # Plotting specific arguments
        # fig args
        figsize: Tuple[Numeric, Numeric] = (12, 8),
        aspect: Numeric = 1.618,
        height: Numeric = 0.8,
        # plot args
        grid: bool = True,
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 12,
        # title args
        title: Optional[str] = None,
        title_fontsize: int = 16,
        title_xadjust: Numeric = 0.5,
        title_yadjust: Numeric = 1.05,
        # legend args
        legend: bool = True,
        legend_labels: Optional[List[str]] = None,
        legend_title: Optional[str] = None,
        legend_loc: Optional[str] = "best",
        legend_fontsize: int = 14,
        legend_ncol: int = 1,
        legend_bbox_to_anchor: Optional[Tuple[Numeric, Numeric]] = None,
        legend_frame: bool = True,
        # return args
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        on_axis: Optional[plt.Axes] = None,
        # args, kwargs
        *args,
        **kwargs,
    ):
        if on_axis:
            fig: plt.Figure = on_axis.get_figure()
            ax: plt.Axes = on_axis
        else:
            fig: plt.Figure
            ax: plt.Axes
            fig, ax = plt.subplots(figsize=figsize)

        dfx: pd.DataFrame = self.df.copy()

        if title is not None:
            title_newline_adjust: float = 0.2
            if title_yadjust is not None and title_yadjust != 1.0:
                title_newline_adjust = abs(title_yadjust - 1.0)
            title_height: float = title.count("\n") * title_newline_adjust
            # increase the title_yadjust by the title_height
            title_yadjust = title_yadjust + title_height

        ax.set_title(
            title,
            fontsize=title_fontsize,
            x=title_xadjust,
            y=title_yadjust,
        )

        if compare_series:
            _cid, _xcat = compare_series.split("_", 1)
            if _cid not in dfx["cid"].unique() or _xcat not in dfx["xcat"].unique():
                raise ValueError(
                    f"Series `{compare_series}` not in DataFrame - used as "
                    "`compare_series`."
                )

            comp_df = (
                dfx.loc[
                    (dfx["cid"] == _cid) & (dfx["xcat"] == _xcat), ["real_date", metric]
                ]
                .copy()
                .reset_index(drop=True)
            )

        valid_tickers = dfx[["cid", "xcat"]].drop_duplicates().values.tolist()

        for xcat in self.xcats:
            for cid in self.cids:
                # Get the unique cid values in dfx for xcat and check if cid is in it
                if [cid, xcat] in valid_tickers:
                    _df = dfx.loc[(dfx["cid"] == cid) & (dfx["xcat"] == xcat), :].copy()
                    _df = _df.sort_values(by="real_date", ascending=True).reset_index(
                        drop=True
                    )
                    ax.plot(_df["real_date"], _df[metric], label=f"{cid}_{xcat}")

        # if there is a compare_series, plot it on the same axis, using a red dashed line
        if compare_series:
            ax.plot(
                comp_df["real_date"],
                comp_df[metric],
                color="red",
                linestyle="--",
                label=compare_series,
            )

        if grid:
            ax.grid(axis="both", linestyle="--", alpha=0.5)

        if x_axis_label:
            ax.set_xlabel(x_axis_label, fontsize=axis_fontsize)

        if y_axis_label:
            ax.set_ylabel(y_axis_label, fontsize=axis_fontsize)

        # if there is a legend, add it
        if legend:
            ax.legend(
                labels=legend_labels if legend_labels else None,
                title=legend_title,
                loc=legend_loc,
                fontsize=legend_fontsize,
                ncol=legend_ncol,
                bbox_to_anchor=legend_bbox_to_anchor,
                frameon=legend_frame,
            )

        plt.tight_layout()
        title: str = title if title else f"LinePlot: Viewing `{metric}`"

        if save_to_file:
            plt.savefig(
                save_to_file,
                dpi=dpi,
                bbox_inches="tight",
            )

        if return_figure:
            return fig

        if show:
            plt.show()
            return


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_test_df
    from macrosynergy.download import LocalCache as JPMaQSDownload

    cids: List[str] = [
        "USD",
        "EUR",
        "GBP",
        "AUD",
        "CAD",
        "JPY",
        "CHF",
        "NZD",
        "SEK",
        "INR",
    ]
    # Quantamental categories of interest

    xcats = [
        "NIR_NSA",
        "RIR_NSA",
        "DU05YXR_NSA",
        "DU05YXR_VT10",
        "FXXR_NSA",
        "EQXR_NSA",
        "DU05YXR_NSA",
    ]  # market links

    sel_cids: List[str] = ["USD", "EUR", "GBP"]
    sel_xcats: List[str] = ["NIR_NSA", "RIR_NSA", "FXXR_NSA", "EQXR_NSA"]

    with JPMaQSDownload(
        local_path=r"~\Macrosynergy\Macrosynergy - Documents\SharedData\JPMaQSTickers"
    ) as jpmaqs:
        df: pd.DataFrame = jpmaqs.download(
            cids=cids,
            xcats=xcats,
            start_date="2016-01-01",
        )

    from random import SystemRandom

    random = SystemRandom()

    # random.seed(42)

    # for cidx, xcatx in df[["cid", "xcat"]].drop_duplicates().values.tolist():
    #     # if random() > 0.5 multiply by random.random()*10
    #     _bools = (df["cid"] == cidx) & (df["xcat"] == xcatx)
    #     r = max(random.random(), 0.1)
    #     df.loc[_bools, "value"] = df.loc[_bools, "value"] * r

    # FacetPlot(df).lineplot()
    import time

    print("From same object:")
    timer_start: float = time.time()

    LinePlot(df, cids=cids, xcats=xcats).plot(
        title=(
            "Test Title with a very long title to see how it looks, \n and a new "
            "line - why not?"
        ),
        legend_fontsize=8,
        compare_series="USD_RIR_NSA",
    )

    # facet_size=(5, 4),
    print(f"Time taken: {time.time() - timer_start}")



"""
Collection of non-standard scikit-learn performance metrics for evaluation of
machine learning model predictions.
"""

import numpy as np
import pandas as pd
import datetime

from statsmodels.regression.mixed_linear_model import MixedLM
from statsmodels.tools.tools import add_constant

from sklearn.metrics import make_scorer, accuracy_score, balanced_accuracy_score
from sklearn.model_selection import cross_val_score
from sklearn.linear_model import LinearRegression

from macrosynergy.learning.panel_time_series_split import ExpandingKFoldPanelSplit

from typing import Union


def panel_significance_probability(
    y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]
) -> float:
    """
    Function to create a linear mixed effects model between the ground truth returns and
    the predicted returns, returning the significance of the model slope.
    Period-specific random effects are included in the model to account for
    return cross-sectional correlations.
    This can be passed into sklearn's make_scorer function to be used as a scorer in a
    grid search or cross validation procedure.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float> significance_prob: 1 - p-value of the regression slope parameter,
        given by the linear mixed effects model.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    if np.all(y_true == 0):
        # Sklearn averages each metric over the CV splits.
        # If all the ground truth labels are zero, the regression is invalid due to a
        # singular matrix. Hence, we return zero in this case.
        significance_prob = 0
        return significance_prob

    # regress ground truth against predictions
    X = add_constant(y_pred)
    groups = y_true.index.get_level_values(1)

    # fit model
    re = MixedLM(y_true, X, groups=groups).fit(reml=False)
    pval = re.pvalues.iloc[1]

    return 1 - pval


def regression_accuracy(y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]) -> float:
    """
    Function to return the accuracy between the signs of the predictions and targets.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Accuracy between the signs of the predictions and targets.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    return accuracy_score(y_true < 0, y_pred < 0)


def regression_balanced_accuracy(
    y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]
) -> float:
    """
    Function to return the balanced accuracy between the signs
    of the predictions and targets.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Balanced accuracy between the signs of the predictions and targets.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    return balanced_accuracy_score(y_true < 0, y_pred < 0)


def sharpe_ratio(y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]) -> float:
    """
    Function to return a Sharpe ratio for a strategy where we go long if the predictions
    are positive and short if the predictions are negative.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Sharpe ratio for the binary strategy.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    portfolio_returns = np.where(y_pred > 0, y_true, -y_true)
    average_return = np.mean(portfolio_returns)
    std_return = np.std(portfolio_returns)

    if std_return == 0:
        # Sklearn averages each metric over the CV splits.
        # If the standard deviation is zero, the portfolio returns are constant.
        # So we return zero to ignore this split in the average.
        sharpe_ratio = 0
    else:
        sharpe_ratio = average_return / std_return

    return sharpe_ratio


def sortino_ratio(y_true: pd.Series, y_pred: Union[pd.Series, np.ndarray]) -> float:
    """
    Function to return a Sortino ratio for a strategy where we go long if the predictions
    are positive and short if the predictions are negative.

    :param <pd.Series> y_true: Pandas series of ground truth labels. These must be
        multi-indexed by cross-section and date. The dates must be in datetime format.
    :param <Union[pd.Series,np.ndarray]> y_pred: Either a pandas series or numpy array
        of predicted targets. This must have the same length as y_true.

    :return <float>: Sortino ratio for the binary strategy.
    """

    # checks
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pandas series")

    if not isinstance(y_true.index, pd.MultiIndex):
        raise ValueError("y_true must be multi-indexed.")

    if not isinstance(y_true.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")

    if not (len(y_true) == len(y_pred)):
        raise ValueError("y_true and y_pred must have the same length.")

    portfolio_returns = np.where(y_pred > 0, y_true, -y_true)
    negative_returns = portfolio_returns[portfolio_returns < 0]
    average_return = np.mean(portfolio_returns)
    denominator = np.sqrt(np.mean(negative_returns**2))

    if denominator == 0:
        # Sklearn averages each metric over the CV splits.
        # If the denominator is zero, the sortino ratio over this period is invalid.
        # So we return zero to ignore this split in the average.
        sortino_ratio = 0
    else:
        sortino_ratio = average_return / denominator

    return sortino_ratio


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CPI", "GROWTH", "RIR"]

    df_cids = pd.DataFrame(index=cids, columns=["earliest", "latest"])
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31"]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31"]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31"]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31"]

    tuples = []

    for cid in cids:
        # get list of all elidgible dates
        sdate = df_cids.loc[cid]["earliest"]
        edate = df_cids.loc[cid]["latest"]
        all_days = pd.date_range(sdate, edate)
        work_days = all_days[all_days.weekday < 5]
        for work_day in work_days:
            tuples.append((cid, work_day))

    n_samples = len(tuples)
    ftrs = np.random.normal(loc=0, scale=1, size=(n_samples, 3))
    labels = np.random.normal(loc=0, scale=1, size=n_samples)
    df = pd.DataFrame(
        data=np.concatenate((np.reshape(labels, (-1, 1)), ftrs), axis=1),
        index=pd.MultiIndex.from_tuples(tuples, names=["cid", "real_date"]),
        columns=xcats,
        dtype=np.float32,
    )

    X = df.drop(columns="XR")
    y = df["XR"]

    splitter = ExpandingKFoldPanelSplit(n_splits=4)
    scorer1 = make_scorer(panel_significance_probability, greater_is_better=True)
    scorer2 = make_scorer(sharpe_ratio, greater_is_better=True)
    scorer3 = make_scorer(sortino_ratio, greater_is_better=True)
    scorer4 = make_scorer(regression_accuracy, greater_is_better=True)
    scorer5 = make_scorer(regression_balanced_accuracy, greater_is_better=True)
    cv_results1 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer1
    )
    cv_results2 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer2
    )
    cv_results3 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer3
    )
    cv_results4 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer4
    )
    cv_results5 = cross_val_score(
        LinearRegression(), X, y, cv=splitter, scoring=scorer5
    )
    print("Probabilities of significances, per split:", cv_results1)
    print("Sharpe ratios, per split:", cv_results2)
    print("Sortino ratios, per split:", cv_results3)
    print("Regression accuracies, per split:", cv_results4)
    print("Regression balanced accuracies, per split:", cv_results5)
    print("Done")


"""
Constants used throughout the package.

- FREQUENCY_MAP:  used to map the frequencies from D, W, M, Q, A to Business Day Frequencies.
"""

FREQUENCY_MAP = {
    "D": "B",
    "W": "W-FRI",
    "M": "BM",
    "Q": "BQ",
    "A": "BA",
}
